@@ -40,9 +40,8 @@
 #include <stdlib.h>
 #include <stdint.h>
 #include <string.h>
-#ifdef HAVE_PWD_H
-# include <pwd.h>
-#endif
+#include <limits.h>
+#include <sys/types.h>
 #include <unistd.h>
 #include <errno.h>
 #ifdef HAVE_STAT
@@ -57,6 +56,10 @@
 # include <sys/time.h>
 # include <sys/resource.h>
 #endif
+#ifdef HAVE_PWD_H
+# include <pwd.h>
+# include <grp.h>
+#endif /*HAVE_PWD_H*/
 #ifdef HAVE_W32_SYSTEM
 # if WINVER < 0x0500
 #   define WINVER 0x0500  /* Required for AllowSetForegroundWindow.  */
@@ -107,6 +110,11 @@ struct gnupg_dir_s
  * an explanation of these file names.  */
 static int allow_special_filenames;
 
+#ifdef HAVE_W32_SYSTEM
+/* State of gnupg_inhibit_set_foregound_window.  */
+static int inhibit_set_foregound_window;
+#endif
+
 
 static GPGRT_INLINE gpg_error_t
 my_error_from_syserror (void)
@@ -538,7 +546,7 @@ translate_sys2libc_fd (gnupg_fd_t fd, int for_write)
 
   /* Note that _open_osfhandle is currently defined to take and return
      a long.  */
-  x = _open_osfhandle ((long)fd, for_write ? 1 : 0);
+  x = _open_osfhandle ((intptr_t)fd, for_write ? 1 : 0);
   if (x == -1)
     log_error ("failed to translate osfhandle %p\n", (void *) fd);
   return x;
@@ -669,7 +677,7 @@ gnupg_tmpfile (void)
           int fd = (int)file;
           fp = _wfdopen (fd, L"w+b");
 #else
-          int fd = _open_osfhandle ((long)file, 0);
+          int fd = _open_osfhandle ((intptr_t)file, 0);
           if (fd == -1)
             {
               CloseHandle (file);
@@ -770,6 +778,20 @@ gnupg_reopen_std (const char *pgmname)
 }
 
 
+/* Inhibit calls to AllowSetForegroundWindow on Windows.  Calling this
+ * with YES set to true calls to gnupg_allow_set_foregound_window are
+ * shunted.  */
+void
+gnupg_inhibit_set_foregound_window (int yes)
+{
+#ifdef HAVE_W32_SYSTEM
+  inhibit_set_foregound_window = yes;
+#else
+  (void)yes;
+#endif
+}
+
+
 /* Hack required for Windows.  */
 void
 gnupg_allow_set_foregound_window (pid_t pid)
@@ -778,6 +800,8 @@ gnupg_allow_set_foregound_window (pid_t pid)
     log_info ("%s called with invalid pid %lu\n",
               "gnupg_allow_set_foregound_window", (unsigned long)pid);
 #if defined(HAVE_W32_SYSTEM) && !defined(HAVE_W32CE_SYSTEM)
+  else if (inhibit_set_foregound_window)
+    ;
   else if (!AllowSetForegroundWindow ((pid_t)pid == (pid_t)(-1)?ASFW_ANY:pid))
     log_info ("AllowSetForegroundWindow(%lu) failed: %s\n",
                (unsigned long)pid, w32_strerror (-1));
@@ -963,35 +987,9 @@ modestr_to_mode (const char *modestr, mode_t oldmode)
 int
 gnupg_mkdir (const char *name, const char *modestr)
 {
-#if GPG_ERROR_VERSION_NUMBER < 0x011c00 /* 1.28 */
- #ifdef HAVE_W32CE_SYSTEM
-  wchar_t *wname;
-  (void)modestr;
-
-  wname = utf8_to_wchar (name);
-  if (!wname)
-    return -1;
-  if (!CreateDirectoryW (wname, NULL))
-    {
-      xfree (wname);
-      return -1;  /* ERRNO is automagically provided by gpg-error.h.  */
-    }
-  xfree (wname);
-  return 0;
- #elif MKDIR_TAKES_ONE_ARG
-  (void)modestr;
-  /* Note: In the case of W32 we better use CreateDirectory and try to
-     set appropriate permissions.  However using mkdir is easier
-     because this sets ERRNO.  */
-  return mkdir (name);
- #else
-  return mkdir (name, modestr_to_mode (modestr, 0));
- #endif
-#else
   /* Note that gpgrt_mkdir also sets ERRNO in addition to returing an
    * gpg-error style error code.  */
   return gpgrt_mkdir (name, modestr);
-#endif
 }
 
 
@@ -1000,40 +998,9 @@ gnupg_mkdir (const char *name, const char *modestr)
 int
 gnupg_chdir (const char *name)
 {
-#if GPG_ERROR_VERSION_NUMBER < 0x011c00 /* 1.28 */
-  return chdir (name);
-#else /* Use the improved version from libgpg_error.  */
-  /* Note that gpgrt_chdir also sets ERRNO in addition to returning a
+  /* Note that gpgrt_chdir also sets ERRNO in addition to returning an
    * gpg-error style error code.  */
   return gpgrt_chdir (name);
-#endif
-}
-
-
-/* A wrapper around rmdir.  NAME is expected to be utf8 encoded.  */
-int
-gnupg_rmdir (const char *name)
-{
-#ifdef HAVE_W32_SYSTEM
-  int rc;
-  wchar_t *wfname;
-
-  wfname = utf8_to_wchar (name);
-  if (!wfname)
-    rc = 0;
-  else
-    {
-      rc = RemoveDirectoryW (wfname);
-      if (!rc)
-        gnupg_w32_set_errno (-1);
-      xfree (wfname);
-    }
-  if (!rc)
-    return -1;
-  return 0;
-#else
-  return rmdir (name);
-#endif
 }
 
 
@@ -1257,53 +1224,7 @@ gnupg_unsetenv (const char *name)
 char *
 gnupg_getcwd (void)
 {
-#if GPGRT_VERSION_NUMBER < 0x012800 /* 1.40 */
-# ifdef HAVE_W32_SYSTEM
-  wchar_t wbuffer[MAX_PATH + sizeof(wchar_t)];
-  DWORD wlen;
-  char *buf, *p;
-
-  wlen = GetCurrentDirectoryW (MAX_PATH, wbuffer);
-  if (!wlen)
-    {
-      gpg_err_set_errno (EINVAL);
-      return NULL;
-
-    }
-  else if (wlen > MAX_PATH)
-    {
-      gpg_err_set_errno (ENAMETOOLONG);
-      return NULL;
-    }
-  buf = wchar_to_utf8 (wbuffer);
-  if (buf)
-    {
-      for (p=buf; *p; p++)
-        if (*p == '\\')
-          *p = '/';
-    }
-  return buf;
-
-# else /*Unix*/
-  char *buffer;
-  size_t size = 100;
-
-  for (;;)
-    {
-      buffer = xtrymalloc (size+1);
-      if (!buffer)
-        return NULL;
-      if (getcwd (buffer, size) == buffer)
-        return buffer;
-      xfree (buffer);
-      if (errno != ERANGE)
-        return NULL;
-      size *= 2;
-    }
-# endif /*Unix*/
-#else
   return gpgrt_getcwd ();
-#endif
 }
 
 
@@ -1312,26 +1233,7 @@ gnupg_getcwd (void)
 gpg_err_code_t
 gnupg_access (const char *name, int mode)
 {
-#if GPGRT_VERSION_NUMBER < 0x012800 /* 1.40 */
-# ifdef HAVE_W32_SYSTEM
-  wchar_t *wfname;
-  gpg_err_code_t ec;
-
-  wfname = utf8_to_wchar (name);
-  if (!wfname)
-    ec = gpg_err_code_from_syserror ();
-  else
-    {
-      ec = _waccess (wfname, mode)? gpg_err_code_from_syserror () : 0;
-      xfree (wfname);
-    }
-  return ec;
-# else
-  return access (name, mode)? gpg_err_code_from_syserror () : 0;
-# endif
-#else /* gpgrt 1.40 or newer.  */
   return gpgrt_access (name, mode);
-#endif
 }
 
 
@@ -1377,55 +1279,6 @@ gnupg_stat (const char *name, struct stat *statbuf)
 #endif /*HAVE_STAT*/
 
 
-/* Wrapper around fopen for the cases where we have not yet switched
- * to es_fopen.  Note that for convenience the prototype is in util.h */
-FILE *
-gnupg_fopen (const char *fname, const char *mode)
-{
-#ifdef HAVE_W32_SYSTEM
-  if (any8bitchar (fname))
-    {
-      wchar_t *wfname;
-      const wchar_t *wmode;
-      wchar_t *wmodebuf = NULL;
-      FILE *ret;
-
-      wfname = utf8_to_wchar (fname);
-      if (!wfname)
-        return NULL;
-      if (!strcmp (mode, "r"))
-        wmode = L"r";
-      else if (!strcmp (mode, "rb"))
-        wmode = L"rb";
-      else if (!strcmp (mode, "w"))
-        wmode = L"w";
-      else if (!strcmp (mode, "wb"))
-        wmode = L"wb";
-      else
-        {
-          wmodebuf = utf8_to_wchar (mode);
-          if (!wmodebuf)
-            {
-              xfree (wfname);
-              return NULL;
-            }
-          wmode = wmodebuf;
-        }
-      ret = _wfopen (wfname, wmode);
-      xfree (wfname);
-      xfree (wmodebuf);
-      return ret;
-    }
-  else
-    return fopen (fname, mode);
-
-#else /*Unix*/
-  return fopen (fname, mode);
-#endif /*Unix*/
-}
-
-
-
 /* A wrapper around open to handle Unicode file names under Windows.  */
 int
 gnupg_open (const char *name, int flags, unsigned int mode)
@@ -1596,6 +1449,100 @@ gnupg_closedir (gnupg_dir_t gdir)
 }
 
 
+/* Try to set an envvar.  Print only a notice on error.  */
+#ifndef HAVE_W32_SYSTEM
+static void
+try_set_envvar (const char *name, const char *value, int silent)
+{
+  if (gnupg_setenv (name, value, 1))
+    if (!silent)
+      log_info ("error setting envvar %s to '%s': %s\n", name, value,
+                gpg_strerror (my_error_from_syserror ()));
+}
+#endif /*!HAVE_W32_SYSTEM*/
+
+
+/* Switch to USER which is either a name or an UID.  This is a nop
+ * under Windows.  Note that in general it is only possible to switch
+ * to another user id if the process is running under root.  if silent
+ * is set no diagnostics are printed.  */
+gpg_error_t
+gnupg_chuid (const char *user, int silent)
+{
+#ifdef HAVE_W32_SYSTEM
+  (void)user;  /* Not implemented for Windows - ignore.  */
+  (void)silent;
+  return 0;
+
+#elif HAVE_PWD_H /* A proper Unix  */
+  unsigned long ul;
+  struct passwd *pw;
+  struct stat st;
+  char *endp;
+  gpg_error_t err;
+
+  gpg_err_set_errno (0);
+  ul = strtoul (user, &endp, 10);
+  if (errno || endp == user || *endp)
+    pw = getpwnam (user);  /* Not a number; assume USER is a name.  */
+  else
+    pw = getpwuid ((uid_t)ul);
+
+  if (!pw)
+    {
+      if (!silent)
+        log_error ("user '%s' not found\n", user);
+      return my_error (GPG_ERR_NOT_FOUND);
+    }
+
+  /* Try to set some envvars even if we are already that user.  */
+  if (!stat (pw->pw_dir, &st))
+    try_set_envvar ("HOME", pw->pw_dir, silent);
+
+  try_set_envvar ("USER", pw->pw_name, silent);
+  try_set_envvar ("LOGNAME", pw->pw_name, silent);
+#ifdef _AIX
+  try_set_envvar ("LOGIN", pw->pw_name, silent);
+#endif
+
+  if (getuid () == pw->pw_uid)
+    return 0;  /* We are already this user.  */
+
+  /* If we need to switch set PATH to a standard value and make sure
+   * GNUPGHOME is not set. */
+  try_set_envvar ("PATH", "/usr/local/bin:/usr/bin:/bin", silent);
+  if (gnupg_unsetenv ("GNUPGHOME"))
+    if (!silent)
+      log_info ("error unsetting envvar %s: %s\n", "GNUPGHOME",
+                gpg_strerror (gpg_error_from_syserror ()));
+
+  if (initgroups (pw->pw_name, pw->pw_gid))
+    {
+      err = my_error_from_syserror ();
+      if (!silent)
+        log_error ("error setting supplementary groups for '%s': %s\n",
+                   pw->pw_name, gpg_strerror (err));
+      return err;
+    }
+
+  if (setuid (pw->pw_uid))
+    {
+      err = my_error_from_syserror ();
+      log_error ("error switching to user '%s': %s\n",
+                 pw->pw_name, gpg_strerror (err));
+      return err;
+    }
+
+  return 0;
+
+#else /*!HAVE_PWD_H */
+  if (!silent)
+    log_info ("system is missing passwd querying functions\n");
+  return my_error (GPG_ERR_NOT_IMPLEMENTED);
+#endif
+}
+
+
 
 #ifdef HAVE_W32CE_SYSTEM
 /* There is a isatty function declaration in cegcc but it does not
@@ -1899,49 +1846,3 @@ gnupg_fd_valid (int fd)
   close (d);
   return 1;
 }
-
-
-/* Return a malloced copy of the current user's account name; this may
- * return NULL on memory failure.  Note that this should eventually be
- * replaced by a gpgrt function. */
-char *
-gnupg_getusername (void)
-{
-  char *result = NULL;
-
-#ifdef HAVE_W32_SYSTEM
-  wchar_t wtmp[1];
-  wchar_t *wbuf;
-  DWORD wsize = 1;
-
-  GetUserNameW (wtmp, &wsize);
-  wbuf = xtrymalloc (wsize * sizeof *wbuf);
-  if (!wbuf)
-    {
-      gpg_err_set_errno (ENOMEM);
-      return NULL;
-    }
-  if (!GetUserNameW (wbuf, &wsize))
-    {
-      gpg_err_set_errno (EINVAL);
-      xfree (wbuf);
-      return NULL;
-    }
-  result= wchar_to_utf8 (wbuf);
-  xfree (wbuf);
-
-#else /* !HAVE_W32_SYSTEM */
-
-# if defined(HAVE_PWD_H) && defined(HAVE_GETPWUID)
-  struct passwd *pwd;
-
-  pwd = getpwuid (getuid());
-  if (pwd)
-    result = xtrystrdup (pwd->pw_name);
-
-# endif /*HAVE_PWD_H*/
-
-#endif /* !HAVE_W32_SYSTEM */
-
-  return result;
-}
