@@ -28,17 +28,28 @@
 # include <getopt.h>
 #endif
 #include <stdlib.h>
-#include <npth.h>
+#include <errno.h>
+#include <assert.h>
 
+#ifdef _WIN32
+# include <winsock2.h>
+# include <winldap.h>
+#else
+# ifdef NEED_LBER_H
+#  include <lber.h>
+# endif
+/* For OpenLDAP, to enable the API that we're using. */
+# define LDAP_DEPRECATED 1
+# include <ldap.h>
+#endif
+#include <npth.h>
 
 #include "dirmngr.h"
 #include "misc.h"
 #include "../common/userids.h"
 #include "../common/mbox-util.h"
 #include "ks-engine.h"
-#include "ldap-misc.h"
 #include "ldap-parse-uri.h"
-#include "ldapserver.h"
 
 
 /* Flags with infos from the connected server.  */
@@ -48,11 +59,177 @@
 #define SERVERINFO_NTDS     8 /* Server is an Active Directory.      */
 
 
+
 #ifndef HAVE_TIMEGM
 time_t timegm(struct tm *tm);
 #endif
+
+/* Convert an LDAP error to a GPG error.  */
+static int
+ldap_err_to_gpg_err (int code)
+{
+  gpg_err_code_t ec;
 
+  switch (code)
+    {
+#ifdef LDAP_X_CONNECTING
+    case LDAP_X_CONNECTING: ec = GPG_ERR_LDAP_X_CONNECTING; break;
+#endif
 
+    case LDAP_REFERRAL_LIMIT_EXCEEDED: ec = GPG_ERR_LDAP_REFERRAL_LIMIT; break;
+    case LDAP_CLIENT_LOOP: ec = GPG_ERR_LDAP_CLIENT_LOOP; break;
+    case LDAP_NO_RESULTS_RETURNED: ec = GPG_ERR_LDAP_NO_RESULTS; break;
+    case LDAP_CONTROL_NOT_FOUND: ec = GPG_ERR_LDAP_CONTROL_NOT_FOUND; break;
+    case LDAP_NOT_SUPPORTED: ec = GPG_ERR_LDAP_NOT_SUPPORTED; break;
+    case LDAP_CONNECT_ERROR: ec = GPG_ERR_LDAP_CONNECT; break;
+    case LDAP_NO_MEMORY: ec = GPG_ERR_LDAP_NO_MEMORY; break;
+    case LDAP_PARAM_ERROR: ec = GPG_ERR_LDAP_PARAM; break;
+    case LDAP_USER_CANCELLED: ec = GPG_ERR_LDAP_USER_CANCELLED; break;
+    case LDAP_FILTER_ERROR: ec = GPG_ERR_LDAP_FILTER; break;
+    case LDAP_AUTH_UNKNOWN: ec = GPG_ERR_LDAP_AUTH_UNKNOWN; break;
+    case LDAP_TIMEOUT: ec = GPG_ERR_LDAP_TIMEOUT; break;
+    case LDAP_DECODING_ERROR: ec = GPG_ERR_LDAP_DECODING; break;
+    case LDAP_ENCODING_ERROR: ec = GPG_ERR_LDAP_ENCODING; break;
+    case LDAP_LOCAL_ERROR: ec = GPG_ERR_LDAP_LOCAL; break;
+    case LDAP_SERVER_DOWN: ec = GPG_ERR_LDAP_SERVER_DOWN; break;
+
+    case LDAP_SUCCESS: ec = GPG_ERR_LDAP_SUCCESS; break;
+
+    case LDAP_OPERATIONS_ERROR: ec = GPG_ERR_LDAP_OPERATIONS; break;
+    case LDAP_PROTOCOL_ERROR: ec = GPG_ERR_LDAP_PROTOCOL; break;
+    case LDAP_TIMELIMIT_EXCEEDED: ec = GPG_ERR_LDAP_TIMELIMIT; break;
+    case LDAP_SIZELIMIT_EXCEEDED: ec = GPG_ERR_LDAP_SIZELIMIT; break;
+    case LDAP_COMPARE_FALSE: ec = GPG_ERR_LDAP_COMPARE_FALSE; break;
+    case LDAP_COMPARE_TRUE: ec = GPG_ERR_LDAP_COMPARE_TRUE; break;
+    case LDAP_AUTH_METHOD_NOT_SUPPORTED: ec=GPG_ERR_LDAP_UNSUPPORTED_AUTH;break;
+    case LDAP_STRONG_AUTH_REQUIRED: ec = GPG_ERR_LDAP_STRONG_AUTH_RQRD; break;
+    case LDAP_PARTIAL_RESULTS: ec = GPG_ERR_LDAP_PARTIAL_RESULTS; break;
+    case LDAP_REFERRAL: ec = GPG_ERR_LDAP_REFERRAL; break;
+
+#ifdef LDAP_ADMINLIMIT_EXCEEDED
+    case LDAP_ADMINLIMIT_EXCEEDED: ec = GPG_ERR_LDAP_ADMINLIMIT; break;
+#endif
+
+#ifdef LDAP_UNAVAILABLE_CRITICAL_EXTENSION
+    case LDAP_UNAVAILABLE_CRITICAL_EXTENSION:
+                               ec = GPG_ERR_LDAP_UNAVAIL_CRIT_EXTN; break;
+#endif
+
+    case LDAP_CONFIDENTIALITY_REQUIRED: ec = GPG_ERR_LDAP_CONFIDENT_RQRD; break;
+    case LDAP_SASL_BIND_IN_PROGRESS: ec = GPG_ERR_LDAP_SASL_BIND_INPROG; break;
+    case LDAP_NO_SUCH_ATTRIBUTE: ec = GPG_ERR_LDAP_NO_SUCH_ATTRIBUTE; break;
+    case LDAP_UNDEFINED_TYPE: ec = GPG_ERR_LDAP_UNDEFINED_TYPE; break;
+    case LDAP_INAPPROPRIATE_MATCHING: ec = GPG_ERR_LDAP_BAD_MATCHING; break;
+    case LDAP_CONSTRAINT_VIOLATION: ec = GPG_ERR_LDAP_CONST_VIOLATION; break;
+
+#ifdef LDAP_TYPE_OR_VALUE_EXISTS
+    case LDAP_TYPE_OR_VALUE_EXISTS: ec = GPG_ERR_LDAP_TYPE_VALUE_EXISTS; break;
+#endif
+
+    case LDAP_INVALID_SYNTAX: ec = GPG_ERR_LDAP_INV_SYNTAX; break;
+    case LDAP_NO_SUCH_OBJECT: ec = GPG_ERR_LDAP_NO_SUCH_OBJ; break;
+    case LDAP_ALIAS_PROBLEM: ec = GPG_ERR_LDAP_ALIAS_PROBLEM; break;
+    case LDAP_INVALID_DN_SYNTAX: ec = GPG_ERR_LDAP_INV_DN_SYNTAX; break;
+    case LDAP_IS_LEAF: ec = GPG_ERR_LDAP_IS_LEAF; break;
+    case LDAP_ALIAS_DEREF_PROBLEM: ec = GPG_ERR_LDAP_ALIAS_DEREF; break;
+
+#ifdef LDAP_X_PROXY_AUTHZ_FAILURE
+    case LDAP_X_PROXY_AUTHZ_FAILURE: ec = GPG_ERR_LDAP_X_PROXY_AUTH_FAIL; break;
+#endif
+
+    case LDAP_INAPPROPRIATE_AUTH: ec = GPG_ERR_LDAP_BAD_AUTH; break;
+    case LDAP_INVALID_CREDENTIALS: ec = GPG_ERR_LDAP_INV_CREDENTIALS; break;
+
+#ifdef LDAP_INSUFFICIENT_ACCESS
+    case LDAP_INSUFFICIENT_ACCESS: ec = GPG_ERR_LDAP_INSUFFICIENT_ACC; break;
+#endif
+
+    case LDAP_BUSY: ec = GPG_ERR_LDAP_BUSY; break;
+    case LDAP_UNAVAILABLE: ec = GPG_ERR_LDAP_UNAVAILABLE; break;
+    case LDAP_UNWILLING_TO_PERFORM: ec = GPG_ERR_LDAP_UNWILL_TO_PERFORM; break;
+    case LDAP_LOOP_DETECT: ec = GPG_ERR_LDAP_LOOP_DETECT; break;
+    case LDAP_NAMING_VIOLATION: ec = GPG_ERR_LDAP_NAMING_VIOLATION; break;
+    case LDAP_OBJECT_CLASS_VIOLATION: ec = GPG_ERR_LDAP_OBJ_CLS_VIOLATION; break;
+    case LDAP_NOT_ALLOWED_ON_NONLEAF: ec=GPG_ERR_LDAP_NOT_ALLOW_NONLEAF;break;
+    case LDAP_NOT_ALLOWED_ON_RDN: ec = GPG_ERR_LDAP_NOT_ALLOW_ON_RDN; break;
+    case LDAP_ALREADY_EXISTS: ec = GPG_ERR_LDAP_ALREADY_EXISTS; break;
+    case LDAP_NO_OBJECT_CLASS_MODS: ec = GPG_ERR_LDAP_NO_OBJ_CLASS_MODS; break;
+    case LDAP_RESULTS_TOO_LARGE: ec = GPG_ERR_LDAP_RESULTS_TOO_LARGE; break;
+    case LDAP_AFFECTS_MULTIPLE_DSAS: ec = GPG_ERR_LDAP_AFFECTS_MULT_DSAS; break;
+
+#ifdef LDAP_VLV_ERROR
+    case LDAP_VLV_ERROR: ec = GPG_ERR_LDAP_VLV; break;
+#endif
+
+    case LDAP_OTHER: ec = GPG_ERR_LDAP_OTHER; break;
+
+#ifdef LDAP_CUP_RESOURCES_EXHAUSTED
+    case LDAP_CUP_RESOURCES_EXHAUSTED: ec=GPG_ERR_LDAP_CUP_RESOURCE_LIMIT;break;
+    case LDAP_CUP_SECURITY_VIOLATION: ec=GPG_ERR_LDAP_CUP_SEC_VIOLATION; break;
+    case LDAP_CUP_INVALID_DATA: ec = GPG_ERR_LDAP_CUP_INV_DATA; break;
+    case LDAP_CUP_UNSUPPORTED_SCHEME: ec = GPG_ERR_LDAP_CUP_UNSUP_SCHEME; break;
+    case LDAP_CUP_RELOAD_REQUIRED: ec = GPG_ERR_LDAP_CUP_RELOAD; break;
+#endif
+
+#ifdef LDAP_CANCELLED
+    case LDAP_CANCELLED: ec = GPG_ERR_LDAP_CANCELLED; break;
+#endif
+
+#ifdef LDAP_NO_SUCH_OPERATION
+    case LDAP_NO_SUCH_OPERATION: ec = GPG_ERR_LDAP_NO_SUCH_OPERATION; break;
+#endif
+
+#ifdef LDAP_TOO_LATE
+    case LDAP_TOO_LATE: ec = GPG_ERR_LDAP_TOO_LATE; break;
+#endif
+
+#ifdef LDAP_CANNOT_CANCEL
+    case LDAP_CANNOT_CANCEL: ec = GPG_ERR_LDAP_CANNOT_CANCEL; break;
+#endif
+
+#ifdef LDAP_ASSERTION_FAILED
+    case LDAP_ASSERTION_FAILED: ec = GPG_ERR_LDAP_ASSERTION_FAILED; break;
+#endif
+
+#ifdef LDAP_PROXIED_AUTHORIZATION_DENIED
+    case LDAP_PROXIED_AUTHORIZATION_DENIED:
+                                      ec = GPG_ERR_LDAP_PROX_AUTH_DENIED; break;
+#endif
+
+    default:
+#if defined(LDAP_E_ERROR) && defined(LDAP_X_ERROR)
+      if (LDAP_E_ERROR (code))
+        ec = GPG_ERR_LDAP_E_GENERAL;
+      else if (LDAP_X_ERROR (code))
+        ec = GPG_ERR_LDAP_X_GENERAL;
+      else
+#endif
+        ec = GPG_ERR_LDAP_GENERAL;
+      break;
+    }
+
+  return ec;
+}
+
+/* Retrieve an LDAP error and return it's GPG equivalent.  */
+static int
+ldap_to_gpg_err (LDAP *ld)
+{
+#if defined(HAVE_LDAP_GET_OPTION) && defined(LDAP_OPT_ERROR_NUMBER)
+  int err;
+
+  if (ldap_get_option (ld, LDAP_OPT_ERROR_NUMBER, &err) == 0)
+    return ldap_err_to_gpg_err (err);
+  else
+    return GPG_ERR_GENERAL;
+#elif defined(HAVE_LDAP_LD_ERRNO)
+  return ldap_err_to_gpg_err (ld->ld_errno);
+#else
+  /* We should never get here since the LDAP library should always
+     have either ldap_get_option or ld_errno, but just in case... */
+  return GPG_ERR_INTERNAL;
+#endif
+}
 
 static time_t
 ldap2epochtime (const char *timestr)
@@ -118,16 +295,6 @@ epoch2ldaptime (time_t stamp)
     return xstrdup ("INVALID TIME");
 }
 #endif
-
-
-static void
-my_ldap_value_free (char **vals)
-{
-  if (vals)
-    ldap_value_free (vals);
-}
-
-
 
 /* Print a help output for the schemata supported by this module. */
 gpg_error_t
@@ -135,40 +302,32 @@ ks_ldap_help (ctrl_t ctrl, parsed_uri_t uri)
 {
   const char data[] =
     "Handler for LDAP URLs:\n"
-    "  ldap://HOST:PORT/[BASEDN]????[bindname=BINDNAME,password=PASSWORD]\n"
+    "  ldap://host:port/[BASEDN]???[bindname=BINDNAME,password=PASSWORD]\n"
     "\n"
     "Note: basedn, bindname and password need to be percent escaped. In\n"
     "particular, spaces need to be replaced with %20 and commas with %2c.\n"
-    "Thus bindname will typically be of the form:\n"
+    "bindname will typically be of the form:\n"
     "\n"
     "  uid=user%2cou=PGP%20Users%2cdc=EXAMPLE%2cdc=ORG\n"
     "\n"
     "The ldaps:// and ldapi:// schemes are also supported.  If ldaps is used\n"
     "then the server's certificate will be checked.  If it is not valid, any\n"
-    "operation will be aborted.  Note that ldaps means LDAP with STARTTLS\n"
-    "\n"
-    "As an alternative to an URL a string in this form may be used:\n"
-    "\n"
-    "  HOST:PORT:BINDNAME:PASSWORD:BASEDN:FLAGS:\n"
-    "\n"
-    "The use of the percent sign or a colon in one of the string values is\n"
-    "currently not supported.\n"
+    "operation will be aborted.\n"
     "\n"
     "Supported methods: search, get, put\n";
   gpg_error_t err;
 
   if(!uri)
     err = ks_print_help (ctrl, "  ldap");
-  else if (!strcmp (uri->scheme, "ldap")
-           || !strcmp (uri->scheme, "ldaps")
-           || !strcmp (uri->scheme, "ldapi")
-           || uri->opaque)
+  else if (uri->is_ldap)
     err = ks_print_help (ctrl, data);
   else
     err = 0;
 
   return err;
 }
+
+
 
 /* Convert a keyspec to a filter.  Return an error if the keyspec is
    bad or is not supported.  The filter is escaped and returned in
@@ -220,8 +379,7 @@ keyspec_to_ldap_filter (const char *keyspec, char **filter, int only_exact,
       else
         p = freeme;
       if ((serverinfo & SERVERINFO_SCHEMAV2))
-        f = xasprintf ("(&(gpgMailbox=%s)(!(|(pgpRevoked=1)(pgpDisabled=1))))",
-                       p);
+        f = xasprintf ("(gpgMailbox=%s)", p);
       else if (!only_exact)
         f = xasprintf ("(pgpUserID=*<%s>*)", p);
       break;
@@ -246,12 +404,10 @@ keyspec_to_ldap_filter (const char *keyspec, char **filter, int only_exact,
 		     (ulong) desc.u.kid[0], (ulong) desc.u.kid[1]);
       break;
 
-    case KEYDB_SEARCH_MODE_FPR16:
-    case KEYDB_SEARCH_MODE_FPR20:
     case KEYDB_SEARCH_MODE_FPR:
       if ((serverinfo & SERVERINFO_SCHEMAV2))
         {
-          freeme = bin2hex (desc.u.fpr, 20, NULL);
+          freeme = bin2hex (desc.u.fpr, desc.fprlen, NULL);
           if (!freeme)
             return gpg_error_from_syserror ();
           f = xasprintf ("(|(gpgFingerprint=%s)(gpgSubFingerprint=%s))",
@@ -324,176 +480,96 @@ keyspec_to_ldap_filter (const char *keyspec, char **filter, int only_exact,
    If no LDAP error occurred, you still need to check that *basednp is
    valid.  If it is NULL, then the server does not appear to be an
    OpenPGP Keyserver.  */
-static gpg_error_t
+static int
 my_ldap_connect (parsed_uri_t uri, LDAP **ldap_connp,
-                 char **r_basedn, char **r_host, int *r_use_tls,
-                 unsigned int *r_serverinfo)
+                 char **basednp, unsigned int *r_serverinfo)
 {
-  gpg_error_t err = 0;
-  int lerr;
-  ldap_server_t server = NULL;
+  int err = 0;
   LDAP *ldap_conn = NULL;
+  char *user = uri->auth;
+  struct uri_tuple_s *password_param;
+  char *password;
   char *basedn = NULL;
-  char *host = NULL;   /* Host to use.  */
-  int port;            /* Port to use.  */
-  int use_tls;         /* 1 = starttls, 2 = ldap-over-tls  */
-  int use_ntds;        /* Use Active Directory authentication.  */
-  const char *bindname;
-  const char *password;
-  const char *basedn_arg;
-#ifndef HAVE_W32_SYSTEM
-  char *tmpstr;
-#endif
 
-  if (r_basedn)
-    *r_basedn = NULL;
-  if (r_host)
-    *r_host = NULL;
-  if (r_use_tls)
-    *r_use_tls = 0;
   *r_serverinfo = 0;
 
-  if (uri->opaque)
-    {
-      server = ldapserver_parse_one (uri->path, NULL, 0);
-      if (!server)
-        return gpg_error (GPG_ERR_LDAP_OTHER);
-      host = server->host;
-      port = server->port;
-      bindname = server->user;
-      password = bindname? server->pass : NULL;
-      basedn_arg = server->base;
-      use_tls = server->starttls? 1 : server->ldap_over_tls? 2 : 0;
-      use_ntds = server->ntds;
-    }
-  else
-    {
-      host = uri->host;
-      port = uri->port;
-      bindname = uri->auth;
-      password = bindname? uri_query_value (uri, "password") : NULL;
-      basedn_arg = uri->path;
-      use_tls = uri->use_tls ? 1 : 0;
-      use_ntds = uri->ad_current;
-    }
-
-  if (!port)
-    port = use_tls == 2? 636 : 389;
-
-
-  if (host)
-    {
-      host = xtrystrdup (host);
-      if (!host)
-        {
-          err = gpg_error_from_syserror ();
-          goto out;
-        }
-    }
-
-  if (opt.verbose)
-    log_info ("ldap connect to '%s:%d:%s:%s:%s:%s%s'\n",
-              host, port,
-              basedn_arg ? basedn_arg : "",
-              bindname ? bindname : "",
-              password ? "*****" : "",
-              use_tls == 1? "starttls" : use_tls == 2? "ldaptls" : "plain",
-              use_ntds ? ",ntds":"");
+  password_param = uri_query_lookup (uri, "password");
+  password = password_param ? password_param->value : NULL;
 
+  if (opt.debug)
+    log_debug ("my_ldap_connect(%s:%d/%s????%s%s%s%s%s%s)\n",
+               uri->host, uri->port,
+               uri->path ? uri->path : "",
+               uri->auth ? "bindname=" : "",
+               uri->auth ? uri->auth : "",
+               uri->auth && password ? "," : "",
+               password ? "password=" : "",
+               password ? ">not shown<": "",
+               uri->ad_current? " auth=>current_user<":"");
 
   /* If the uri specifies a secure connection and we don't support
      TLS, then fail; don't silently revert to an insecure
      connection.  */
-  if (use_tls)
+  if (uri->use_tls)
     {
 #ifndef HAVE_LDAP_START_TLS_S
-      log_error ("ldap: can't connect to the server: no TLS support.");
+      log_error ("Can't use LDAP to connect to the server: no TLS support.");
       err = GPG_ERR_LDAP_NOT_SUPPORTED;
       goto out;
 #endif
     }
 
-
-#ifdef HAVE_W32_SYSTEM
-  /* Note that host==NULL uses the default domain controller.  */
-  npth_unprotect ();
-  ldap_conn = ldap_sslinit (host, port, (use_tls == 2));
-  npth_protect ();
+  ldap_conn = ldap_init (uri->host, uri->port);
   if (!ldap_conn)
     {
-      lerr = LdapGetLastError ();
-      err = ldap_err_to_gpg_err (lerr);
-      log_error ("error initializing LDAP '%s:%d': %s\n",
-                 host, port, ldap_err2string (lerr));
+      err = gpg_err_code_from_syserror ();
+      log_error ("error initializing LDAP for (%s://%s:%d)\n",
+                 uri->scheme, uri->host, uri->port);
       goto out;
     }
-#else /* Unix */
-  tmpstr = xtryasprintf ("%s://%s:%d",
-                         use_tls == 2? "ldaps" : "ldap",
-                         host, port);
-  if (!tmpstr)
-    {
-      err = gpg_error_from_syserror ();
-      goto out;
-    }
-  npth_unprotect ();
-  lerr = ldap_initialize (&ldap_conn, tmpstr);
-  npth_protect ();
-  if (lerr != LDAP_SUCCESS || !ldap_conn)
-    {
-      err = ldap_err_to_gpg_err (lerr);
-      log_error ("error initializing LDAP '%s': %s\n",
-                 tmpstr, ldap_err2string (lerr));
-      xfree (tmpstr);
-      goto out;
-    }
-  xfree (tmpstr);
-#endif /* Unix */
 
 #ifdef HAVE_LDAP_SET_OPTION
   {
     int ver = LDAP_VERSION3;
 
-    lerr = ldap_set_option (ldap_conn, LDAP_OPT_PROTOCOL_VERSION, &ver);
-    if (lerr != LDAP_SUCCESS)
+    err = ldap_set_option (ldap_conn, LDAP_OPT_PROTOCOL_VERSION, &ver);
+    if (err != LDAP_SUCCESS)
       {
 	log_error ("ks-ldap: unable to go to LDAP 3: %s\n",
-		   ldap_err2string (lerr));
-	err = ldap_err_to_gpg_err (lerr);
+		   ldap_err2string (err));
 	goto out;
       }
   }
-  if (opt.ldaptimeout)
-    {
-      int ver = opt.ldaptimeout;
-
-      lerr = ldap_set_option (ldap_conn, LDAP_OPT_TIMELIMIT, &ver);
-      if (lerr != LDAP_SUCCESS)
-        {
-          log_error ("ks-ldap: unable to set LDAP timelimit to %us: %s\n",
-                     opt.ldaptimeout, ldap_err2string (lerr));
-          err = ldap_err_to_gpg_err (lerr);
-          goto out;
-        }
-      if (opt.verbose)
-        log_info ("ldap timeout set to %us\n", opt.ldaptimeout);
-    }
 #endif
 
+  /* XXX: It would be nice to have an option to provide the server's
+     certificate.  */
+#if 0
+#if defined(LDAP_OPT_X_TLS_CACERTFILE) && defined(HAVE_LDAP_SET_OPTION)
+  err = ldap_set_option (NULL, LDAP_OPT_X_TLS_CACERTFILE, ca_cert_file);
+  if (err)
+    {
+      log_error ("unable to set ca-cert-file to '%s': %s\n",
+		 ca_cert_file, ldap_err2string (err));
+      goto out;
+    }
+#endif /* LDAP_OPT_X_TLS_CACERTFILE && HAVE_LDAP_SET_OPTION */
+#endif
 
 #ifdef HAVE_LDAP_START_TLS_S
-  if (use_tls == 1)
+  if (uri->use_tls)
     {
+      /* XXX: We need an option to determine whether to abort if the
+	 certificate is bad or not.  Right now we conservatively
+	 default to checking the certificate and aborting.  */
 #ifndef HAVE_W32_SYSTEM
       int check_cert = LDAP_OPT_X_TLS_HARD; /* LDAP_OPT_X_TLS_NEVER */
 
-      lerr = ldap_set_option (ldap_conn,
-                              LDAP_OPT_X_TLS_REQUIRE_CERT, &check_cert);
-      if (lerr)
+      err = ldap_set_option (ldap_conn,
+			     LDAP_OPT_X_TLS_REQUIRE_CERT, &check_cert);
+      if (err)
 	{
-	  log_error ("ldap: error setting an TLS option: %s\n",
-                     ldap_err2string (lerr));
-          err = ldap_err_to_gpg_err (lerr);
+	  log_error ("error setting TLS option on LDAP connection\n");
 	  goto out;
 	}
 #else
@@ -505,7 +581,7 @@ my_ldap_connect (parsed_uri_t uri, LDAP **ldap_connp,
 #endif
 
       npth_unprotect ();
-      lerr = ldap_start_tls_s (ldap_conn,
+      err = ldap_start_tls_s (ldap_conn,
 #ifdef HAVE_W32_SYSTEM
 			      /* ServerReturnValue, result */
 			      NULL, NULL,
@@ -513,44 +589,45 @@ my_ldap_connect (parsed_uri_t uri, LDAP **ldap_connp,
 			      /* ServerControls, ClientControls */
 			      NULL, NULL);
       npth_protect ();
-      if (lerr)
+      if (err)
 	{
-	  log_error ("ldap: error switching to STARTTLS mode: %s\n",
-                     ldap_err2string (lerr));
-          err = ldap_err_to_gpg_err (lerr);
+	  log_error ("error connecting to LDAP server with TLS\n");
 	  goto out;
 	}
     }
 #endif
 
-  if (use_ntds)
+  if (uri->ad_current)
     {
+      if (opt.debug)
+        log_debug ("LDAP bind to current user via AD\n");
 #ifdef HAVE_W32_SYSTEM
       npth_unprotect ();
-      lerr = ldap_bind_s (ldap_conn, NULL, NULL, LDAP_AUTH_NEGOTIATE);
+      err = ldap_bind_s (ldap_conn, NULL, NULL, LDAP_AUTH_NEGOTIATE);
       npth_protect ();
-      if (lerr != LDAP_SUCCESS)
+      if (err != LDAP_SUCCESS)
 	{
 	  log_error ("error binding to LDAP via AD: %s\n",
-                     ldap_err2string (lerr));
-          err = ldap_err_to_gpg_err (lerr);
+                     ldap_err2string (err));
 	  goto out;
 	}
 #else
-      log_error ("ldap: no Active Directory support but 'ntds' requested\n");
       err = gpg_error (GPG_ERR_NOT_SUPPORTED);
       goto out;
 #endif
     }
-  else if (bindname)
+  else if (uri->auth)
     {
+      if (opt.debug)
+        log_debug ("LDAP bind to %s, password %s\n",
+                   user, password ? ">not shown<" : ">none<");
+
       npth_unprotect ();
-      lerr = ldap_simple_bind_s (ldap_conn, bindname, password);
+      err = ldap_simple_bind_s (ldap_conn, user, password);
       npth_protect ();
-      if (lerr != LDAP_SUCCESS)
+      if (err != LDAP_SUCCESS)
 	{
-	  log_error ("error binding to LDAP: %s\n", ldap_err2string (lerr));
-          err = ldap_err_to_gpg_err (lerr);
+	  log_error ("error binding to LDAP: %s\n", ldap_err2string (err));
 	  goto out;
 	}
     }
@@ -559,16 +636,13 @@ my_ldap_connect (parsed_uri_t uri, LDAP **ldap_connp,
       /* By default we don't bind as there is usually no need to.  */
     }
 
-  if (basedn_arg && *basedn_arg)
+  if (uri->path && *uri->path)
     {
-      /* User specified base DN.  In this case we know the server is a
+      /* User specified base DN.  */
+      basedn = xstrdup (uri->path);
+
+      /* If the user specifies a base DN, then we know the server is a
        * real LDAP server.  */
-      basedn = xtrystrdup (basedn_arg);
-      if (!basedn)
-        {
-          err = gpg_error_from_syserror ();
-          goto out;
-        }
       *r_serverinfo |= SERVERINFO_REALLDAP;
     }
   else
@@ -577,11 +651,11 @@ my_ldap_connect (parsed_uri_t uri, LDAP **ldap_connp,
       char *attr[] = { "namingContexts", NULL };
 
       npth_unprotect ();
-      lerr = ldap_search_s (ldap_conn, "", LDAP_SCOPE_BASE,
+      err = ldap_search_s (ldap_conn, "", LDAP_SCOPE_BASE,
 			   "(objectClass=*)", attr, 0, &res);
       npth_protect ();
 
-      if (lerr == LDAP_SUCCESS)
+      if (err == LDAP_SUCCESS)
 	{
 	  char **context;
 
@@ -600,7 +674,7 @@ my_ldap_connect (parsed_uri_t uri, LDAP **ldap_connp,
 
               *r_serverinfo |= SERVERINFO_REALLDAP;
 
-	      for (i = 0; context[i] && !basedn; i++)
+	      for (i = 0; context[i] && ! basedn; i++)
 		{
 		  char **vals;
 		  LDAPMessage *si_res;
@@ -610,42 +684,42 @@ my_ldap_connect (parsed_uri_t uri, LDAP **ldap_connp,
                     char *object = xasprintf ("cn=pgpServerInfo,%s",
                                               context[i]);
                     npth_unprotect ();
-                    lerr = ldap_search_s (ldap_conn, object, LDAP_SCOPE_BASE,
+                    err = ldap_search_s (ldap_conn, object, LDAP_SCOPE_BASE,
                                          "(objectClass=*)", attr2, 0, &si_res);
                     npth_protect ();
                     xfree (object);
                   }
 
-		  if (lerr == LDAP_SUCCESS)
+		  if (err == LDAP_SUCCESS)
 		    {
 		      vals = ldap_get_values (ldap_conn, si_res,
 					      "pgpBaseKeySpaceDN");
-		      if (vals && vals[0])
+		      if (vals)
 			{
 			  basedn = xtrystrdup (vals[0]);
+			  ldap_value_free (vals);
 			}
-                      my_ldap_value_free (vals);
 
 		      vals = ldap_get_values (ldap_conn, si_res,
 					      "pgpSoftware");
-		      if (vals && vals[0])
+		      if (vals)
 			{
                           if (opt.debug)
                             log_debug ("Server: \t%s\n", vals[0]);
                           if (!ascii_strcasecmp (vals[0], "GnuPG"))
                             is_gnupg = 1;
+			  ldap_value_free (vals);
 			}
-                      my_ldap_value_free (vals);
 
 		      vals = ldap_get_values (ldap_conn, si_res,
 					      "pgpVersion");
-		      if (vals && vals[0])
+		      if (vals)
 			{
                           if (opt.debug)
                             log_debug ("Version:\t%s\n", vals[0]);
                           if (is_gnupg)
                             {
-                              char *fields[2];
+                              const char *fields[2];
                               int nfields;
                               nfields = split_fields (vals[0],
                                                       fields, DIM(fields));
@@ -655,8 +729,8 @@ my_ldap_connect (parsed_uri_t uri, LDAP **ldap_connp,
                                   && !ascii_strcasecmp (fields[1], "ntds"))
                                 *r_serverinfo |= SERVERINFO_NTDS;
                             }
+			  ldap_value_free (vals);
 			}
-                      my_ldap_value_free (vals);
 		    }
 
 		  /* From man ldap_search_s: "res parameter of
@@ -669,7 +743,7 @@ my_ldap_connect (parsed_uri_t uri, LDAP **ldap_connp,
 	      ldap_value_free (context);
 	    }
 	}
-      else /* ldap_search failed.  */
+      else
 	{
 	  /* We don't have an answer yet, which means the server might
 	     be a PGP.com keyserver. */
@@ -679,32 +753,32 @@ my_ldap_connect (parsed_uri_t uri, LDAP **ldap_connp,
 	  char *attr2[] = { "pgpBaseKeySpaceDN", "version", "software", NULL };
 
           npth_unprotect ();
-	  lerr = ldap_search_s (ldap_conn, "cn=pgpServerInfo", LDAP_SCOPE_BASE,
+	  err = ldap_search_s (ldap_conn, "cn=pgpServerInfo", LDAP_SCOPE_BASE,
 			       "(objectClass=*)", attr2, 0, &si_res);
           npth_protect ();
-	  if (lerr == LDAP_SUCCESS)
+	  if (err == LDAP_SUCCESS)
 	    {
 	      /* For the PGP LDAP keyserver, this is always
 	       * "OU=ACTIVE,O=PGP KEYSPACE,C=US", but it might not be
                * in the future. */
 
 	      vals = ldap_get_values (ldap_conn, si_res, "baseKeySpaceDN");
-	      if (vals && vals[0])
+	      if (vals)
 		{
 		  basedn = xtrystrdup (vals[0]);
+		  ldap_value_free (vals);
 		}
-              my_ldap_value_free (vals);
 
 	      vals = ldap_get_values (ldap_conn, si_res, "software");
-	      if (vals && vals[0])
+	      if (vals)
 		{
                   if (opt.debug)
                     log_debug ("ks-ldap: PGP Server: \t%s\n", vals[0]);
+		  ldap_value_free (vals);
 		}
-              my_ldap_value_free (vals);
 
 	      vals = ldap_get_values (ldap_conn, si_res, "version");
-	      if (vals && vals[0])
+	      if (vals)
 		{
                   if (opt.debug)
                     log_debug ("ks-ldap: PGP Server Version:\t%s\n", vals[0]);
@@ -718,8 +792,8 @@ my_ldap_connect (parsed_uri_t uri, LDAP **ldap_connp,
 		  if (atoi (vals[0]) > 1)
                     *r_serverinfo |= SERVERINFO_PGPKEYV2;
 
+		  ldap_value_free (vals);
 		}
-              my_ldap_value_free (vals);
 	    }
 
 	  ldap_msgfree (si_res);
@@ -742,27 +816,23 @@ my_ldap_connect (parsed_uri_t uri, LDAP **ldap_connp,
                  (*r_serverinfo & SERVERINFO_PGPKEYV2)? "pgpKeyV2":"pgpKey");
     }
 
-  ldapserver_list_free (server);
+  if (err)
+    xfree (basedn);
+  else
+    {
+      if (basednp)
+	*basednp = basedn;
+      else
+	xfree (basedn);
+    }
 
   if (err)
     {
-      xfree (basedn);
       if (ldap_conn)
 	ldap_unbind (ldap_conn);
     }
   else
-    {
-      if (r_basedn)
-	*r_basedn = basedn;
-      else
-	xfree (basedn);
-      if (r_host)
-        *r_host = host;
-      else
-        xfree (host);
-
-      *ldap_connp = ldap_conn;
-    }
+    *ldap_connp = ldap_conn;
 
   return err;
 }
@@ -777,17 +847,10 @@ extract_keys (estream_t output,
   char **vals;
 
   es_fprintf (output, "INFO %s BEGIN\n", certid);
+  es_fprintf (output, "pub:%s:", certid);
 
   /* Note: ldap_get_values returns a NULL terminated array of
      strings.  */
-
-  vals = ldap_get_values (ldap_conn, message, "gpgfingerprint");
-  if (vals && vals[0] && vals[0][0])
-    es_fprintf (output, "pub:%s:", vals[0]);
-  else
-    es_fprintf (output, "pub:%s:", certid);
-  my_ldap_value_free (vals);
-
   vals = ldap_get_values (ldap_conn, message, "pgpkeytype");
   if (vals && vals[0])
     {
@@ -795,8 +858,8 @@ extract_keys (estream_t output,
 	es_fprintf  (output, "1");
       else if (strcmp (vals[0],"DSS/DH") == 0)
 	es_fprintf (output, "17");
+      ldap_value_free (vals);
     }
-  my_ldap_value_free (vals);
 
   es_fprintf (output, ":");
 
@@ -806,8 +869,8 @@ extract_keys (estream_t output,
       int v = atoi (vals[0]);
       if (v > 0)
 	es_fprintf (output, "%d", v);
+      ldap_value_free (vals);
     }
-  my_ldap_value_free (vals);
 
   es_fprintf (output, ":");
 
@@ -816,8 +879,8 @@ extract_keys (estream_t output,
     {
       if (strlen (vals[0]) == 15)
 	es_fprintf (output, "%u", (unsigned int) ldap2epochtime (vals[0]));
+      ldap_value_free (vals);
     }
-  my_ldap_value_free (vals);
 
   es_fprintf (output, ":");
 
@@ -826,8 +889,8 @@ extract_keys (estream_t output,
     {
       if (strlen (vals[0]) == 15)
 	es_fprintf (output, "%u", (unsigned int) ldap2epochtime (vals[0]));
+      ldap_value_free (vals);
     }
-  my_ldap_value_free (vals);
 
   es_fprintf (output, ":");
 
@@ -836,8 +899,8 @@ extract_keys (estream_t output,
     {
       if (atoi (vals[0]) == 1)
 	es_fprintf (output, "r");
+      ldap_value_free (vals);
     }
-  my_ldap_value_free (vals);
 
   es_fprintf (output, "\n");
 
@@ -847,8 +910,8 @@ extract_keys (estream_t output,
       int i;
       for (i = 0; vals[i]; i++)
 	es_fprintf (output, "uid:%s\n", vals[i]);
+      ldap_value_free (vals);
     }
-  my_ldap_value_free (vals);
 
   es_fprintf (output, "INFO %s END\n", certid);
 }
@@ -863,8 +926,6 @@ ks_ldap_get (ctrl_t ctrl, parsed_uri_t uri, const char *keyspec,
   gpg_error_t err = 0;
   int ldap_err;
   unsigned int serverinfo;
-  char *host = NULL;
-  int use_tls;
   char *filter = NULL;
   LDAP *ldap_conn = NULL;
   char *basedn = NULL;
@@ -881,11 +942,12 @@ ks_ldap_get (ctrl_t ctrl, parsed_uri_t uri, const char *keyspec,
     }
 
   /* Make sure we are talking to an OpenPGP LDAP server.  */
-  err = my_ldap_connect (uri, &ldap_conn,
-                         &basedn, &host, &use_tls, &serverinfo);
-  if (err || !basedn)
+  ldap_err = my_ldap_connect (uri, &ldap_conn, &basedn, &serverinfo);
+  if (ldap_err || !basedn)
     {
-      if (!err)
+      if (ldap_err)
+	err = ldap_err_to_gpg_err (ldap_err);
+      else
 	err = gpg_error (GPG_ERR_GENERAL);
       goto out;
     }
@@ -908,7 +970,6 @@ ks_ldap_get (ctrl_t ctrl, parsed_uri_t uri, const char *keyspec,
 	"dummy",
 	"pgpcertid", "pgpuserid", "pgpkeyid", "pgprevoked", "pgpdisabled",
 	"pgpkeycreatetime", "modifytimestamp", "pgpkeysize", "pgpkeytype",
-        "gpgfingerprint",
 	NULL
       };
     /* 1 if we want just attribute types; 0 if we want both attribute
@@ -1011,7 +1072,7 @@ ks_ldap_get (ctrl_t ctrl, parsed_uri_t uri, const char *keyspec,
 		}
 	    }
 
-          my_ldap_value_free (certid);
+	  ldap_value_free (certid);
 	}
 
       free_strlist (seen);
@@ -1021,8 +1082,7 @@ ks_ldap_get (ctrl_t ctrl, parsed_uri_t uri, const char *keyspec,
 
       if (!err && anykey)
         err = dirmngr_status_printf (ctrl, "SOURCE", "%s://%s",
-                                     use_tls? "ldaps" : "ldap",
-                                     host? host:"");
+                                     uri->scheme, uri->host? uri->host:"");
     }
   }
 
@@ -1044,7 +1104,6 @@ ks_ldap_get (ctrl_t ctrl, parsed_uri_t uri, const char *keyspec,
     }
 
   xfree (basedn);
-  xfree (host);
 
   if (ldap_conn)
     ldap_unbind (ldap_conn);
@@ -1079,10 +1138,12 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
     }
 
   /* Make sure we are talking to an OpenPGP LDAP server.  */
-  err = my_ldap_connect (uri, &ldap_conn, &basedn, NULL, NULL, &serverinfo);
-  if (err || !basedn)
+  ldap_err = my_ldap_connect (uri, &ldap_conn, &basedn, &serverinfo);
+  if (ldap_err || !basedn)
     {
-      if (!err)
+      if (ldap_err)
+	err = ldap_err_to_gpg_err (ldap_err);
+      else
 	err = GPG_ERR_GENERAL;
       goto out;
     }
@@ -1116,8 +1177,7 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
       {
 	"pgpcertid", "pgpuserid", "pgprevoked", "pgpdisabled",
 	"pgpkeycreatetime", "pgpkeyexpiretime", "modifytimestamp",
-	"pgpkeysize", "pgpkeytype", "gpgfingerprint",
-        NULL
+	"pgpkeysize", "pgpkeytype", NULL
       };
 
     if (opt.debug)
@@ -1157,7 +1217,6 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
 	    add_to_strlist (&dupelist, certid[0]);
 	    count++;
 	  }
-        my_ldap_value_free (certid);
       }
 
     if (ldap_err == LDAP_SIZELIMIT_EXCEEDED)
@@ -1187,26 +1246,18 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
 	    LDAPMessage *uids;
 
 	    certid = ldap_get_values (ldap_conn, each, "pgpcertid");
-	    if (!certid || !certid[0])
-              {
-                my_ldap_value_free (certid);
-                continue;
-              }
+	    if (! certid || ! certid[0])
+	      continue;
 
 	    /* Have we seen this certid before? */
 	    if (! strlist_find (dupelist, certid[0]))
 	      {
 		add_to_strlist (&dupelist, certid[0]);
 
-                vals = ldap_get_values (ldap_conn, each, "gpgfingerprint");
-                if (vals && vals[0] && vals[0][0])
-                  es_fprintf (fp, "pub:%s:", vals[0]);
-                else
-                  es_fprintf (fp, "pub:%s:", certid[0]);
-                my_ldap_value_free (vals);
+		es_fprintf (fp, "pub:%s:",certid[0]);
 
 		vals = ldap_get_values (ldap_conn, each, "pgpkeytype");
-		if (vals && vals[0])
+		if (vals)
 		  {
 		    /* The LDAP server doesn't exactly handle this
 		       well. */
@@ -1214,60 +1265,60 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
 		      es_fputs ("1", fp);
 		    else if (strcasecmp (vals[0], "DSS/DH") == 0)
 		      es_fputs ("17", fp);
+		    ldap_value_free (vals);
 		  }
-		my_ldap_value_free (vals);
 
 		es_fputc (':', fp);
 
 		vals = ldap_get_values (ldap_conn, each, "pgpkeysize");
-		if (vals && vals[0])
+		if (vals)
 		  {
 		    /* Not sure why, but some keys are listed with a
 		       key size of 0.  Treat that like an unknown. */
 		    if (atoi (vals[0]) > 0)
 		      es_fprintf (fp, "%d", atoi (vals[0]));
+		    ldap_value_free (vals);
 		  }
-                my_ldap_value_free (vals);
 
 		es_fputc (':', fp);
 
 		/* YYYYMMDDHHmmssZ */
 
 		vals = ldap_get_values (ldap_conn, each, "pgpkeycreatetime");
-		if(vals && vals[0] && strlen (vals[0]) == 15)
+		if(vals && strlen (vals[0]) == 15)
 		  {
 		    es_fprintf (fp, "%u",
 				(unsigned int) ldap2epochtime(vals[0]));
+		    ldap_value_free (vals);
 		  }
-                my_ldap_value_free (vals);
 
 		es_fputc (':', fp);
 
 		vals = ldap_get_values (ldap_conn, each, "pgpkeyexpiretime");
-		if (vals && vals[0] && strlen (vals[0]) == 15)
+		if (vals && strlen (vals[0]) == 15)
 		  {
 		    es_fprintf (fp, "%u",
 				(unsigned int) ldap2epochtime (vals[0]));
+		    ldap_value_free (vals);
 		  }
-                my_ldap_value_free (vals);
 
 		es_fputc (':', fp);
 
 		vals = ldap_get_values (ldap_conn, each, "pgprevoked");
-		if (vals && vals[0])
+		if (vals)
 		  {
 		    if (atoi (vals[0]) == 1)
 		      es_fprintf (fp, "r");
+		    ldap_value_free (vals);
 		  }
-                my_ldap_value_free (vals);
 
 		vals = ldap_get_values (ldap_conn, each, "pgpdisabled");
-		if (vals && vals[0])
+		if (vals)
 		  {
 		    if (atoi (vals[0]) ==1)
 		      es_fprintf (fp, "d");
+		    ldap_value_free (vals);
 		  }
-                my_ldap_value_free (vals);
 
 #if 0
 		/* This is not yet specified in the keyserver
@@ -1275,12 +1326,12 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
 		es_fputc (':', fp);
 
 		vals = ldap_get_values (ldap_conn, each, "modifytimestamp");
-		if(vals && vals[0] strlen (vals[0]) == 15)
+		if(vals && strlen (vals[0]) == 15)
 		  {
 		    es_fprintf (fp, "%u",
 				(unsigned int) ldap2epochtime (vals[0]));
+		    ldap_value_free (vals);
 		  }
-                my_ldap_value_free (vals);
 #endif
 
 		es_fprintf (fp, "\n");
@@ -1291,13 +1342,10 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
 		     uids = ldap_next_entry (ldap_conn, uids))
 		  {
 		    vals = ldap_get_values (ldap_conn, uids, "pgpcertid");
-                    if (!vals || !vals[0])
-                      {
-                        my_ldap_value_free (vals);
-                        continue;
-                      }
+		    if (! vals)
+		      continue;
 
-		    if (!ascii_strcasecmp (certid[0], vals[0]))
+		    if (strcasecmp (certid[0], vals[0]) == 0)
 		      {
 			char **uidvals;
 
@@ -1307,14 +1355,12 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
 						   uids, "pgpuserid");
 			if (uidvals)
 			  {
-			    /* Need to percent escape any colons */
-                            char *quoted = try_percent_escape (uidvals[0],
-                                                               NULL);
-                            if (quoted)
-                              es_fputs (quoted, fp);
+			    /* Need to escape any colons */
+			    char *quoted = percent_escape (uidvals[0], NULL);
+			    es_fputs (quoted, fp);
 			    xfree (quoted);
+			    ldap_value_free (uidvals);
 			  }
-                        my_ldap_value_free (uidvals);
 
 			es_fprintf (fp, "\n");
 		      }
@@ -1323,7 +1369,7 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
 		  }
 	      }
 
-            my_ldap_value_free (certid);
+	      ldap_value_free (certid);
 	  }
       }
 
@@ -1337,7 +1383,8 @@ ks_ldap_search (ctrl_t ctrl, parsed_uri_t uri, const char *pattern,
  out:
   if (err)
     {
-      es_fclose (fp);
+      if (fp)
+	es_fclose (fp);
     }
   else
     {
@@ -1654,7 +1701,7 @@ uncescape (char *str)
 	  && hexdigitp (str + r + 3))
 	{
 	  int x = hextobyte (&str[r + 2]);
-	  log_assert (0 <= x && x <= 0xff);
+	  assert (0 <= x && x <= 0xff);
 
 	  str[w] = x;
 
@@ -1919,7 +1966,7 @@ extract_attributes (LDAPMod ***modlist, int *extract_state,
 
       uncescape (uid);
       modlist_add (modlist, "pgpUserID", uid);
-      if (schemav2 && (mbox = mailbox_from_userid (uid)))
+      if (schemav2 && (mbox = mailbox_from_userid (uid, 0)))
         {
           modlist_add (modlist, "gpgMailbox", mbox);
           xfree (mbox);
@@ -1970,10 +2017,12 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
       return gpg_error (GPG_ERR_NOT_SUPPORTED);
     }
 
-  err = my_ldap_connect (uri, &ldap_conn, &basedn, NULL, NULL, &serverinfo);
-  if (err || !basedn)
+  ldap_err = my_ldap_connect (uri, &ldap_conn, &basedn, &serverinfo);
+  if (ldap_err || !basedn)
     {
-      if (!err)
+      if (ldap_err)
+	err = ldap_err_to_gpg_err (ldap_err);
+      else
 	err = GPG_ERR_GENERAL;
       goto out;
     }
@@ -2027,8 +2076,8 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
     {
       dump = es_fopen("/tmp/modlist.txt", "w");
       if (! dump)
-	log_error ("failed to open /tmp/modlist.txt: %s\n",
-		   gpg_strerror (gpg_error_from_syserror ()));
+	log_error ("Failed to open /tmp/modlist.txt: %s\n",
+		   strerror (errno));
 
       if (dump)
 	{
