@@ -30,6 +30,22 @@
 #include "options.h"
 
 
+/* Run time check to see whether mpi_copy does not copy the flags
+ * properly.   This was fixed in version 1.8.6.  */
+static int
+is_mpi_copy_broken (void)
+{
+  static char result;
+
+  if (!result)
+    {
+      result = !gcry_check_version ("1.8.6");
+      result |= 0x80;
+    }
+  return (result & 1);
+}
+
+
 /* This is mpi_copy with a fix for opaque MPIs which store a NULL
    pointer.  This will also be fixed in Libggcrypt 1.7.0.  */
 static gcry_mpi_t
@@ -40,6 +56,17 @@ my_mpi_copy (gcry_mpi_t a)
       && !gcry_mpi_get_opaque (a, NULL))
     return NULL;
 
+  if (is_mpi_copy_broken ())
+    {
+      int flag_user2 = a? gcry_mpi_get_flag (a, GCRYMPI_FLAG_USER2) : 0;
+      gcry_mpi_t b;
+
+      b = gcry_mpi_copy (a);
+      if (b && flag_user2)
+        gcry_mpi_set_flag (b, GCRYMPI_FLAG_USER2);
+      return b;
+    }
+
   return gcry_mpi_copy (a);
 }
 
@@ -77,11 +104,6 @@ free_seckey_enc( PKT_signature *sig )
   xfree(sig->hashed);
   xfree(sig->unhashed);
 
-  if (sig->pka_info)
-    {
-      xfree (sig->pka_info->uri);
-      xfree (sig->pka_info);
-    }
   xfree (sig->signers_uid);
 
   xfree(sig);
@@ -224,25 +246,17 @@ copy_public_key (PKT_public_key *d, PKT_public_key *s)
     }
   else
     d->revkey = NULL;
+
+  if (s->serialno)
+    d->serialno = xstrdup (s->serialno);
+  if (s->updateurl)
+    d->updateurl = xstrdup (s->updateurl);
+
   return d;
 }
 
 
 
-static pka_info_t *
-cp_pka_info (const pka_info_t *s)
-{
-  pka_info_t *d = xmalloc (sizeof *s + strlen (s->email));
-
-  d->valid = s->valid;
-  d->checked = s->checked;
-  d->uri = s->uri? xstrdup (s->uri):NULL;
-  memcpy (d->fpr, s->fpr, sizeof s->fpr);
-  strcpy (d->email, s->email);
-  return d;
-}
-
-
 PKT_signature *
 copy_signature( PKT_signature *d, PKT_signature *s )
 {
@@ -258,7 +272,6 @@ copy_signature( PKT_signature *d, PKT_signature *s )
 	for(i=0; i < n; i++ )
 	    d->data[i] = my_mpi_copy( s->data[i] );
     }
-    d->pka_info = s->pka_info? cp_pka_info (s->pka_info) : NULL;
     d->hashed = cp_subpktarea (s->hashed);
     d->unhashed = cp_subpktarea (s->unhashed);
     if (s->signers_uid)
