@@ -25,7 +25,6 @@
 #include <errno.h>
 #include <unistd.h>
 #include <time.h>
-#include <assert.h>
 
 #include "gpgsm.h"
 
@@ -37,6 +36,7 @@
 #include "../common/i18n.h"
 #include "../common/tlv.h"
 #include "../common/compliance.h"
+#include "../common/pkscreening.h"
 
 struct list_external_parm_s
 {
@@ -49,36 +49,6 @@ struct list_external_parm_s
 };
 
 
-/* This table is to map Extended Key Usage OIDs to human readable
-   names.  */
-struct
-{
-  const char *oid;
-  const char *name;
-} key_purpose_map[] = {
-  { "1.3.6.1.5.5.7.3.1",  "serverAuth" },
-  { "1.3.6.1.5.5.7.3.2",  "clientAuth" },
-  { "1.3.6.1.5.5.7.3.3",  "codeSigning" },
-  { "1.3.6.1.5.5.7.3.4",  "emailProtection" },
-  { "1.3.6.1.5.5.7.3.5",  "ipsecEndSystem" },
-  { "1.3.6.1.5.5.7.3.6",  "ipsecTunnel" },
-  { "1.3.6.1.5.5.7.3.7",  "ipsecUser" },
-  { "1.3.6.1.5.5.7.3.8",  "timeStamping" },
-  { "1.3.6.1.5.5.7.3.9",  "ocspSigning" },
-  { "1.3.6.1.5.5.7.3.10", "dvcs" },
-  { "1.3.6.1.5.5.7.3.11", "sbgpCertAAServerAuth" },
-  { "1.3.6.1.5.5.7.3.13", "eapOverPPP" },
-  { "1.3.6.1.5.5.7.3.14", "wlanSSID" },
-
-  { "2.16.840.1.113730.4.1", "serverGatedCrypto.ns" }, /* Netscape. */
-  { "1.3.6.1.4.1.311.10.3.3", "serverGatedCrypto.ms"}, /* Microsoft. */
-
-  { "1.3.6.1.5.5.7.48.1.5", "ocspNoCheck" },
-
-  { NULL, NULL }
-};
-
-
 /* Do not print this extension in the list of extensions.  This is set
    for oids which are already available via ksba functions. */
 #define OID_FLAG_SKIP 1
@@ -86,6 +56,8 @@ struct
 #define OID_FLAG_UTF8 2
 /* The extension can be trnted as a hex string.  */
 #define OID_FLAG_HEX  4
+/* Define if this specififies a key purpose.  */
+#define OID_FLAG_KP   8
 
 /* A table mapping OIDs to a descriptive string. */
 static struct
@@ -143,7 +115,23 @@ static struct
   { "1.3.6.1.5.5.7.1.10", "acProxying" },
   { "1.3.6.1.5.5.7.1.11", "subjectInfoAccess" },
 
+  { "1.3.6.1.5.5.7.3.1",  "serverAuth", OID_FLAG_KP },
+  { "1.3.6.1.5.5.7.3.2",  "clientAuth", OID_FLAG_KP },
+  { "1.3.6.1.5.5.7.3.3",  "codeSigning", OID_FLAG_KP },
+  { "1.3.6.1.5.5.7.3.4",  "emailProtection", OID_FLAG_KP },
+  { "1.3.6.1.5.5.7.3.5",  "ipsecEndSystem", OID_FLAG_KP }, /* historic */
+  { "1.3.6.1.5.5.7.3.6",  "ipsecTunnel", OID_FLAG_KP },    /* historic */
+  { "1.3.6.1.5.5.7.3.7",  "ipsecUser", OID_FLAG_KP },
+  { "1.3.6.1.5.5.7.3.8",  "timeStamping", OID_FLAG_KP },
+  { "1.3.6.1.5.5.7.3.9",  "ocspSigning", OID_FLAG_KP },
+  { "1.3.6.1.5.5.7.3.10", "dvcs", OID_FLAG_KP },
+  { "1.3.6.1.5.5.7.3.11", "sbgpCertAAServerAuth", OID_FLAG_KP },
+  { "1.3.6.1.5.5.7.3.13", "eapOverPPP", OID_FLAG_KP },
+  { "1.3.6.1.5.5.7.3.14", "wlanSSID", OID_FLAG_KP },
+  { "1.3.6.1.5.5.7.3.17", "ipsecIKE", OID_FLAG_KP },       /* rfc-4945 */
+
   { "1.3.6.1.5.5.7.48.1", "ocsp" },
+  { "1.3.6.1.5.5.7.48.1.5", "ocspNoCheck", OID_FLAG_KP },
   { "1.3.6.1.5.5.7.48.2", "caIssuers" },
   { "1.3.6.1.5.5.7.48.3", "timeStamping" },
   { "1.3.6.1.5.5.7.48.5", "caRepository" },
@@ -171,6 +159,7 @@ static struct
   { "2.5.29.35", "authorityKeyIdentifier", OID_FLAG_SKIP},
   { "2.5.29.36", "policyConstraints" },
   { "2.5.29.37", "extKeyUsage", OID_FLAG_SKIP},
+  { "2.5.29.37.0", "anyExtendedKeyUsage", OID_FLAG_KP},
   { "2.5.29.46", "freshestCRL" },
   { "2.5.29.54", "inhibitAnyPolicy" },
 
@@ -186,11 +175,16 @@ static struct
   { "2.16.840.1.113730.1.11", "netscape-userPicture" },
   { "2.16.840.1.113730.1.12", "netscape-ssl-server-name" },
   { "2.16.840.1.113730.1.13", "netscape-comment" },
+  { "2.16.840.1.113730.4.1", "serverGatedCrypto.ns", OID_FLAG_KP },
 
   /* GnuPG extensions */
   { "1.3.6.1.4.1.11591.2.1.1", "pkaAddress" },
   { "1.3.6.1.4.1.11591.2.2.1", "standaloneCertificate" },
   { "1.3.6.1.4.1.11591.2.2.2", "wellKnownPrivateKey" },
+  { "1.3.6.1.4.1.11591.2.6.1", "gpgUsageCert", OID_FLAG_KP },
+  { "1.3.6.1.4.1.11591.2.6.2", "gpgUsageSign", OID_FLAG_KP },
+  { "1.3.6.1.4.1.11591.2.6.3", "gpgUsageEncr", OID_FLAG_KP },
+  { "1.3.6.1.4.1.11591.2.6.4", "gpgUsageAuth", OID_FLAG_KP },
 
   /* Extensions used by the Bundesnetzagentur.  */
   { "1.3.6.1.4.1.8301.3.5", "validityModel" },
@@ -201,20 +195,40 @@ static struct
   { "1.3.6.1.4.1.41482.3.8", "yubikey-pin-touch-policy", OID_FLAG_HEX },
   { "1.3.6.1.4.1.41482.3.9", "yubikey-formfactor", OID_FLAG_HEX },
 
+  /* Microsoft extensions.  */
+  { "1.3.6.1.4.1.311.3.10.3.12","ms-old-documentSigning", OID_FLAG_KP },
+  { "1.3.6.1.4.1.311.10.3.3", "ms-serverGatedCrypto", OID_FLAG_KP },
+  { "1.3.6.1.4.1.311.10.3.11","ms-keyRecovery", OID_FLAG_KP },
+  { "1.3.6.1.4.1.311.10.3.12","ms-documentSigning", OID_FLAG_KP },
+  { "1.3.6.1.4.1.311.10.3.4", "ms-encryptedFileSystem", OID_FLAG_KP },
+  { "1.3.6.1.4.1.311.10.3.4.1","ms-efsRecovery", OID_FLAG_KP },
+  { "1.3.6.1.4.1.311.20.2.1", "ms-enrollmentAgent", OID_FLAG_KP },
+  { "1.3.6.1.4.1.311.20.2.2", "ms-smartcardLogon", OID_FLAG_KP },
+  { "1.3.6.1.4.1.311.21.5",   "ms-caExchange", OID_FLAG_KP },
+  { "1.3.6.1.4.1.311.21.6",   "ms-keyRecovery", OID_FLAG_KP },
+  { "1.3.6.1.4.1.311.21.19",  "ms-dsEmailReplication", OID_FLAG_KP },
+
+  /* Other vendor extensions.  */
+  { "1.3.6.1.4.1.30205.13.1.1", "trusted-disk", OID_FLAG_KP },
+  { "1.2.840.113583.1.1.5",     "pdfAuthenticDocumentsTrust", OID_FLAG_KP },
+  { "1.3.6.1.4.1.6449.1.3.5.2", "comodoCertifiedDeliveryService", OID_FLAG_KP },
+
   { NULL }
 };
 
 
-/* Return the description for OID; if no description is available
-   NULL is returned. */
+/* Return the description for OID; if no description is available NULL
+ * is returned.  If MATCHFLAG is set the flag of the OID must match
+ * MATCHFLAG; otherwise NULL is returned.  */
 static const char *
-get_oid_desc (const char *oid, unsigned int *flag)
+get_oid_desc (const char *oid, unsigned int matchflag, unsigned int *flag)
 {
   int i;
 
   if (oid)
     for (i=0; oidtranstbl[i].oid; i++)
-      if (!strcmp (oidtranstbl[i].oid, oid))
+      if (!strcmp (oidtranstbl[i].oid, oid)
+          && (!matchflag || (oidtranstbl[i].flag & matchflag)))
         {
           if (flag)
             *flag = oidtranstbl[i].flag;
@@ -246,6 +260,38 @@ print_key_data (ksba_cert_t cert, estream_t fp)
 #endif
 }
 
+
+/* Various public key screenings.  (Right now just ROCA).  With
+ * COLON_MODE set the output is formatted for use in the compliance
+ * field of a colon listing.  */
+static void
+print_pk_screening (ksba_cert_t cert, int colon_mode, estream_t fp)
+{
+  gpg_error_t err;
+  gcry_mpi_t modulus;
+
+  modulus = gpgsm_get_rsa_modulus (cert);
+  if (modulus)
+    {
+      err = screen_key_for_roca (modulus);
+      if (!err)
+        ;
+      else if (gpg_err_code (err) == GPG_ERR_TRUE)
+        {
+          if (colon_mode)
+            es_fprintf (fp, colon_mode > 1? " %d":"%d", 6001);
+          else
+            es_fprintf (fp, "    screening: ROCA vulnerability detected\n");
+        }
+      else if (!colon_mode)
+        es_fprintf (fp, "    screening: [ROCA check failed: %s]\n",
+                    gpg_strerror (err));
+      gcry_mpi_release (modulus);
+    }
+
+}
+
+
 static void
 print_capabilities (ksba_cert_t cert, estream_t fp)
 {
@@ -359,6 +405,7 @@ static void
 print_compliance_flags (ksba_cert_t cert, int algo, unsigned int nbits,
                         estream_t fp)
 {
+  int indent = 0;
   int hashalgo;
 
   /* Note that we do not need to test for PK_ALGO_FLAG_RSAPSS because
@@ -370,8 +417,12 @@ print_compliance_flags (ksba_cert_t cert, int algo, unsigned int nbits,
       if (gnupg_digest_is_compliant (CO_DE_VS, hashalgo))
         {
           es_fputs (gnupg_status_compliance_flag (CO_DE_VS), fp);
+          indent = 1;
         }
     }
+
+  if (opt.with_key_screening)
+    print_pk_screening (cert, 1+indent, fp);
 }
 
 
@@ -390,6 +441,7 @@ list_cert_colon (ctrl_t ctrl, ksba_cert_t cert, unsigned int validity,
   gpg_error_t valerr;
   int algo;
   unsigned int nbits;
+  char *curve = NULL;
   const char *chain_id;
   char *chain_id_buffer = NULL;
   int is_root = 0;
@@ -414,8 +466,9 @@ list_cert_colon (ctrl_t ctrl, ksba_cert_t cert, unsigned int validity,
         chain_id = chain_id_buffer;
         ksba_cert_release (next);
       }
-    else if (rc == -1)  /* We have reached the root certificate. */
+    else if (gpg_err_code (rc) == GPG_ERR_NOT_FOUND)
       {
+        /* We have reached the root certificate. */
         chain_id = fpr;
         is_root = 1;
       }
@@ -481,7 +534,7 @@ list_cert_colon (ctrl_t ctrl, ksba_cert_t cert, unsigned int validity,
   if (*truststring)
     es_fputs (truststring, fp);
 
-  algo = gpgsm_get_key_algo_info (cert, &nbits);
+  algo = gpgsm_get_key_algo_info2 (cert, &nbits, &curve);
   es_fprintf (fp, ":%u:%d:%s:", nbits, algo, fpr+24);
 
   ksba_cert_get_validity (cert, 0, t);
@@ -545,6 +598,8 @@ list_cert_colon (ctrl_t ctrl, ksba_cert_t cert, unsigned int validity,
     }
   es_putc (':', fp);  /* End of field 15. */
   es_putc (':', fp);  /* End of field 16. */
+  if (curve)
+    es_fputs (curve, fp);
   es_putc (':', fp);  /* End of field 17. */
   print_compliance_flags (cert, algo, nbits, fp);
   es_putc (':', fp);  /* End of field 18. */
@@ -559,6 +614,10 @@ list_cert_colon (ctrl_t ctrl, ksba_cert_t cert, unsigned int validity,
   es_putc ('\n', fp);
   xfree (fpr); fpr = NULL; chain_id = NULL;
   xfree (chain_id_buffer); chain_id_buffer = NULL;
+  /* SHA256 FPR record */
+  fpr = gpgsm_get_fingerprint_hexstring (cert, GCRY_MD_SHA256);
+  es_fprintf (fp, "fp2:::::::::%s::::\n", fpr);
+  xfree (fpr); fpr = NULL;
 
   /* Always print the keygrip.  */
   if ( (p = gpgsm_get_keygrip_hexstring (cert)))
@@ -604,6 +663,7 @@ list_cert_colon (ctrl_t ctrl, ksba_cert_t cert, unsigned int validity,
       xfree (p);
     }
   xfree (kludge_uid);
+  xfree (curve);
 }
 
 
@@ -773,6 +833,10 @@ list_cert_raw (ctrl_t ctrl, KEYDB_HANDLE hd,
       es_putc ('\n', fp);
     }
 
+  dn = gpgsm_get_fingerprint_string (cert, GCRY_MD_SHA256);
+  es_fprintf (fp, "     sha2_fpr: %s\n", dn?dn:"error");
+  xfree (dn);
+
   dn = gpgsm_get_fingerprint_string (cert, 0);
   es_fprintf (fp, "     sha1_fpr: %s\n", dn?dn:"error");
   xfree (dn);
@@ -799,16 +863,15 @@ list_cert_raw (ctrl_t ctrl, KEYDB_HANDLE hd,
   es_putc ('\n', fp);
 
   oid = ksba_cert_get_digest_algo (cert);
-  s = get_oid_desc (oid, NULL);
+  s = get_oid_desc (oid, 0, NULL);
   es_fprintf (fp, "     hashAlgo: %s%s%s%s\n", oid, s?" (":"",s?s:"",s?")":"");
 
   {
-    const char *algoname;
-    unsigned int nbits;
+    char *algostr;
 
-    algoname = gcry_pk_algo_name (gpgsm_get_key_algo_info (cert, &nbits));
-    es_fprintf (fp, "      keyType: %u bit %s\n",
-                nbits, algoname? algoname:"?");
+    algostr = gpgsm_pubkey_algo_string (cert, NULL);
+    es_fprintf (fp, "      keyType: %s\n", algostr? algostr : "[error]");
+    xfree (algostr);
   }
 
   /* subjectKeyIdentifier */
@@ -899,10 +962,8 @@ list_cert_raw (ctrl_t ctrl, KEYDB_HANDLE hd,
           while (p && (pend=strchr (p, ':')))
             {
               *pend++ = 0;
-              for (i=0; key_purpose_map[i].oid; i++)
-                if ( !strcmp (key_purpose_map[i].oid, p) )
-                  break;
-              es_fputs (key_purpose_map[i].oid?key_purpose_map[i].name:p, fp);
+              s = get_oid_desc (p, OID_FLAG_KP, NULL);
+              es_fputs (s ? s : p, fp);
               p = pend;
               if (*p != 'C')
                 es_fputs (" (suggested)", fp);
@@ -932,10 +993,8 @@ list_cert_raw (ctrl_t ctrl, KEYDB_HANDLE hd,
           while (p && (pend=strchr (p, ':')))
             {
               *pend++ = 0;
-              for (i=0; key_purpose_map[i].oid; i++)
-                if ( !strcmp (key_purpose_map[i].oid, p) )
-                  break;
-              es_fputs (p, fp);
+              s = get_oid_desc (p, OID_FLAG_KP, NULL);
+              es_fputs (s?s:p, fp);
               p = pend;
               if (*p == 'C')
                 es_fputs (" (critical)", fp);
@@ -1012,7 +1071,7 @@ list_cert_raw (ctrl_t ctrl, KEYDB_HANDLE hd,
                                                          &name)); idx++)
     {
       es_fputs ("     authInfo: ", fp);
-      s = get_oid_desc (string, NULL);
+      s = get_oid_desc (string, 0, NULL);
       es_fprintf (fp, "%s%s%s%s\n", string, s?" (":"", s?s:"", s?")":"");
       print_names_raw (fp, -15, name);
       ksba_name_release (name);
@@ -1029,7 +1088,7 @@ list_cert_raw (ctrl_t ctrl, KEYDB_HANDLE hd,
                                                          &name)); idx++)
     {
       es_fputs ("  subjectInfo: ", fp);
-      s = get_oid_desc (string, NULL);
+      s = get_oid_desc (string, 0, NULL);
       es_fprintf (fp, "%s%s%s%s\n", string, s?" (":"", s?s:"", s?")":"");
       print_names_raw (fp, -15, name);
       ksba_name_release (name);
@@ -1047,7 +1106,7 @@ list_cert_raw (ctrl_t ctrl, KEYDB_HANDLE hd,
     {
       unsigned int flag;
 
-      s = get_oid_desc (oid, &flag);
+      s = get_oid_desc (oid, 0, &flag);
       if ((flag & OID_FLAG_SKIP))
         continue;
 
@@ -1109,12 +1168,12 @@ list_cert_std (ctrl_t ctrl, ksba_cert_t cert, estream_t fp, int have_secret,
   ksba_sexp_t sexp;
   char *dn;
   ksba_isotime_t t;
-  int idx, i;
+  int idx;
   int is_ca, chainlen;
   unsigned int kusage;
   char *string, *p, *pend;
   size_t off, len;
-  const char *oid;
+  const char *oid, *s;
   const unsigned char *cert_der = NULL;
 
 
@@ -1166,15 +1225,13 @@ list_cert_std (ctrl_t ctrl, ksba_cert_t cert, estream_t fp, int have_secret,
 
 
   {
-    const char *algoname;
-    unsigned int nbits;
+    char *algostr;
 
-    algoname = gcry_pk_algo_name (gpgsm_get_key_algo_info (cert, &nbits));
-    es_fprintf (fp, "     key type: %u bit %s\n",
-                nbits, algoname? algoname:"?");
+    algostr = gpgsm_pubkey_algo_string (cert, NULL);
+    es_fprintf (fp, "     key type: %s\n", algostr? algostr : "[error]");
+    xfree (algostr);
   }
 
-
   err = ksba_cert_get_key_usage (cert, &kusage);
   if (gpg_err_code (err) != GPG_ERR_NO_DATA)
     {
@@ -1217,10 +1274,8 @@ list_cert_std (ctrl_t ctrl, ksba_cert_t cert, estream_t fp, int have_secret,
           while (p && (pend=strchr (p, ':')))
             {
               *pend++ = 0;
-              for (i=0; key_purpose_map[i].oid; i++)
-                if ( !strcmp (key_purpose_map[i].oid, p) )
-                  break;
-              es_fputs (key_purpose_map[i].oid?key_purpose_map[i].name:p, fp);
+              s = get_oid_desc (p, OID_FLAG_KP, NULL);
+              es_fputs (s? s : p, fp);
               p = pend;
               if (*p != 'C')
                 es_fputs (" (suggested)", fp);
@@ -1243,7 +1298,7 @@ list_cert_std (ctrl_t ctrl, ksba_cert_t cert, estream_t fp, int have_secret,
         {
           if (!cert_der)
             cert_der = ksba_cert_get_image (cert, NULL);
-          assert (cert_der);
+          log_assert (cert_der);
           es_fputs ("  restriction: ", fp);
           print_utf8_extn (fp, 15, cert_der+off, len);
         }
@@ -1292,7 +1347,11 @@ list_cert_std (ctrl_t ctrl, ksba_cert_t cert, estream_t fp, int have_secret,
     }
 
   dn = gpgsm_get_fingerprint_string (cert, 0);
-  es_fprintf (fp, "  fingerprint: %s\n", dn?dn:"error");
+  es_fprintf (fp, "     sha1 fpr: %s\n", dn?dn:"error");
+  xfree (dn);
+
+  dn = gpgsm_get_fingerprint_string (cert, GCRY_MD_SHA256);
+  es_fprintf (fp, "     sha2 fpr: %s\n", dn?dn:"error");
   xfree (dn);
 
   if (opt.with_keygrip)
@@ -1305,6 +1364,9 @@ list_cert_std (ctrl_t ctrl, ksba_cert_t cert, estream_t fp, int have_secret,
         }
     }
 
+  if (opt.with_key_screening)
+    print_pk_screening (cert, 0, fp);
+
   if (have_secret)
     {
       char *cardsn;
@@ -1341,6 +1403,8 @@ list_cert_std (ctrl_t ctrl, ksba_cert_t cert, estream_t fp, int have_secret,
       else
         es_fprintf (fp, "  [certificate is bad: %s]\n", gpg_strerror (err));
     }
+  if (opt.debug)
+    es_fflush (fp);
 }
 
 
@@ -1393,7 +1457,7 @@ list_internal_keys (ctrl_t ctrl, strlist_t names, estream_t fp,
   int have_secret;
   int want_ephemeral = ctrl->with_ephemeral_keys;
 
-  hd = keydb_new ();
+  hd = keydb_new (ctrl);
   if (!hd)
     {
       log_error ("keydb_new failed\n");
@@ -1445,8 +1509,6 @@ list_internal_keys (ctrl_t ctrl, strlist_t names, estream_t fp,
 
       for (i=0; (i < ndesc
                  && (desc[i].mode == KEYDB_SEARCH_MODE_FPR
-                     || desc[i].mode == KEYDB_SEARCH_MODE_FPR20
-                     || desc[i].mode == KEYDB_SEARCH_MODE_FPR16
                      || desc[i].mode == KEYDB_SEARCH_MODE_KEYGRIP)); i++)
         ;
       if (i == ndesc)
@@ -1550,7 +1612,7 @@ list_internal_keys (ctrl_t ctrl, strlist_t names, estream_t fp,
       lastcert = cert;
       cert = NULL;
     }
-  if (gpg_err_code (rc) == GPG_ERR_EOF || rc == -1 )
+  if (gpg_err_code (rc) == GPG_ERR_NOT_FOUND)
     rc = 0;
   if (rc)
     log_error ("keydb_search failed: %s\n", gpg_strerror (rc));
@@ -1648,3 +1710,98 @@ gpgsm_list_keys (ctrl_t ctrl, strlist_t names, estream_t fp,
     err = list_external_keys (ctrl, names, fp, (mode&256));
   return err;
 }
+
+
+
+static gpg_error_t
+do_show_certs (ctrl_t ctrl, const char *fname, estream_t outfp)
+{
+  gpg_error_t err;
+  gnupg_ksba_io_t b64reader = NULL;
+  ksba_reader_t reader;
+  ksba_cert_t cert = NULL;
+  estream_t fp;
+  int any = 0;
+
+  if (!fname || (fname[0] == '-' && !fname[1]))
+    {
+      fp = es_stdin;
+      fname = "[stdin]";
+    }
+  else
+    {
+      fp = es_fopen (fname, "rb");
+      if (!fp)
+        {
+          err = gpg_error_from_syserror ();
+          log_error (_("can't open '%s': %s\n"), fname, gpg_strerror (err));
+          return err;
+        }
+    }
+
+  err = gnupg_ksba_create_reader
+    (&b64reader, ((ctrl->is_pem? GNUPG_KSBA_IO_PEM : 0)
+                  | (ctrl->is_base64? GNUPG_KSBA_IO_BASE64 : 0)
+                  | (ctrl->autodetect_encoding? GNUPG_KSBA_IO_AUTODETECT : 0)
+                  | GNUPG_KSBA_IO_MULTIPEM),
+     fp, &reader);
+  if (err)
+    {
+      log_error ("can't create reader: %s\n", gpg_strerror (err));
+      goto leave;
+    }
+
+  /* We need to loop here to handle multiple PEM objects per file. */
+  do
+    {
+      ksba_cert_release (cert); cert = NULL;
+
+      err = ksba_cert_new (&cert);
+      if (err)
+        goto leave;
+
+      err = ksba_cert_read_der (cert, reader);
+      if (err)
+        goto leave;
+
+      es_fprintf (outfp, "File ........: %s\n", fname);
+      list_cert_raw (ctrl, NULL, cert, outfp, 0, 0);
+      es_putc ('\n', outfp);
+      any = 1;
+
+      ksba_reader_clear (reader, NULL, NULL);
+    }
+  while (!gnupg_ksba_reader_eof_seen (b64reader));
+
+ leave:
+  if (any && gpg_err_code (err) == GPG_ERR_EOF)
+    err = 0;
+  ksba_cert_release (cert);
+  gnupg_ksba_destroy_reader (b64reader);
+  if (fp != es_stdin)
+    es_fclose (fp);
+  return err;
+}
+
+
+/* Show a raw dump of the certificates found in the files given in
+ * the arrag FILES.  Write output to FP.  */
+gpg_error_t
+gpgsm_show_certs (ctrl_t ctrl, int nfiles, char **files, estream_t fp)
+{
+  gpg_error_t saveerr = 0;
+  gpg_error_t err;
+
+  if (!nfiles)
+    saveerr = do_show_certs (ctrl, NULL, fp);
+  else
+    {
+      for (; nfiles; nfiles--, files++)
+        {
+          err = do_show_certs (ctrl, *files, fp);
+          if (err && !saveerr)
+            saveerr = err;
+        }
+    }
+  return saveerr;
+}
